#include "player_character.hpp"
#include "projectile_pool.hpp"
#include "utils/godot_macros.h"
#include <godot_cpp/classes/navigation_agent3d.hpp>
#include <godot_cpp/classes/time.hpp>
#include <godot_cpp/variant/utility_functions.hpp>

namespace godot {
void PlayerCharacter::_bind_methods() {
#define CLASSNAME PlayerCharacter
    GDPROPERTY_HINTED(rotation_speed_curve, Variant::OBJECT, PROPERTY_HINT_RESOURCE_TYPE, "Curve");
}

void PlayerCharacter::_enter_tree() { GDGAMEONLY();
    this->nav_agent = this->get_node<NavigationAgent3D>("NavigationAgent3D");
    this->target_rotation = this->get_global_transform().get_basis().get_quaternion();
    this->health = this->get_node<Health>("Health");
    this->primary_weapon_pool = this->get_node<ProjectilePool>("ProjectilePool");
}

void PlayerCharacter::_process(double delta_time) { GDGAMEONLY();
    this->process_rotation(delta_time);
    if(!this->mode_manual)
        this->process_ai(delta_time);
    if(this->firing)
        this->try_fire_weapon();
}

void PlayerCharacter::_physics_process(double delta_time) { GDGAMEONLY();
    // accelerate towards velocity target
    Vector3 const new_velocity = this->get_velocity().move_toward(this->velocity_target, delta_time * PlayerCharacter::ACCELERATION);
    // only apply velocity if not grounded
    Vector3 const gravity{this->is_on_floor() ? Vector3() : Vector3{0.f, this->get_velocity().y - 9.8f, 0.f}};
    this->set_velocity(new_velocity + gravity);
    // update position
    this->move_and_slide();
}

void PlayerCharacter::move(Vector3 world_vector) {
    this->velocity_target = world_vector * PlayerCharacter::WALK_SPEED;
}

void PlayerCharacter::aim(Vector3 at) {
    // calculate the forward vector by normalized difference between player character and the target
    Vector3 const position{this->weapon_muzzle->get_global_position()};
    Vector3 const forward{(Vector3{at.x, 0.f, at.z} - Vector3{position.x, 0.f, position.z}).normalized()};
    // we always want up to be the global unit up
    Vector3 const up{0.f, 1.f, 0.f};
    // left is the cross product of the two
    this->target_rotation = Basis{up.cross(forward), up, forward};
}

void PlayerCharacter::set_firing(bool firing) {
    this->firing = firing;
}

void PlayerCharacter::set_manual_mode(bool value) {
    this->mode_manual = value;
    this->nav_agent->set_process_mode(value ? ProcessMode::PROCESS_MODE_PAUSABLE : ProcessMode::PROCESS_MODE_DISABLED);
}

void PlayerCharacter::set_rotation_speed_curve(Ref<Curve> curve) {
    this->rotation_speed_curve = curve;
}

Ref<Curve> PlayerCharacter::get_rotation_speed_curve() const {
    return this->rotation_speed_curve;
}

Health *PlayerCharacter::get_health() {
    return this->health;
}

Health const *PlayerCharacter::get_health() const {
    return this->health;
}

void PlayerCharacter::set_character_data(Ref<CharacterData> data) {
    this->data = data;
    UtilityFunctions::print("PlayerCharacter::set_character_data");
    UtilityFunctions::print(" - data: ", data->get_path());
    UtilityFunctions::print(" - weapon: ", data->get_weapon()->get_path());
    this->fire_interval = 1.f / this->data->get_weapon()->get_rounds_per_second();
    this->primary_weapon_pool->set_data(this->data->get_weapon());
}

void PlayerCharacter::set_weapon_muzzle(Node3D *node) {
    this->weapon_muzzle = node;
}

void PlayerCharacter::process_ai(double delta_time) {
    this->velocity_target = this->nav_agent->get_velocity();
}

void PlayerCharacter::process_rotation(double delta_time) {
    // copy the current transform and basis matrix
    Transform3D trans{this->get_global_transform()};
    Basis basis = trans.get_basis();
    // construct the current rotation ..
    Quaternion const current_quaternion = basis.get_rotation_quaternion();
    // .. and the target rotation from their respective bases
    Quaternion const target_quaternion = this->target_rotation.get_rotation_quaternion();
    // calculate the angle that still needs to be traveled
    float const angle = current_quaternion.angle_to(target_quaternion);
    // calculate the angle amount that can be moved this frame
    float const angle_step{float(this->rotation_speed_curve->sample(angle) * PlayerCharacter::ROTATION_SPEED * delta_time)};
    // update this object's global transform with the new rotation
    basis.set_quaternion(angle < angle_step ? target_quaternion // to avoid overshooting, check if the max step is smaller than the angle distance
                         : current_quaternion.slerp(target_quaternion, angle_step / angle)); // convert the angle step to a lerp t value between current and target rotations
    trans.set_basis(basis);
    this->set_global_transform(trans);
}

void PlayerCharacter::try_fire_weapon() {
    if(float(Time::get_singleton()->get_ticks_msec()) / 1000.f < this->fire_timer)
        return;
    if(!this->data->get_weapon()->get_allow_automatic())
        this->set_firing(false);
    this->fire_timer = float(Time::get_singleton()->get_ticks_msec()) / 1000.f + this->fire_interval;
    Node3D *node = this->primary_weapon_pool->claim_projectile();
    if(node != nullptr)
        node->set_global_transform(this->weapon_muzzle->get_global_transform());
}

float const PlayerCharacter::ACCELERATION{100.f};
float const PlayerCharacter::WALK_SPEED{3.f};
float const PlayerCharacter::SPRINT_SPEED{5.f};
float const PlayerCharacter::ROTATION_SPEED{10.f};
}
